use anyhow::{anyhow, Result};
use std::collections::{BinaryHeap, VecDeque};
use std::fs;

#[derive(Default, Debug)]
enum Ops {
    Plus(u64),
    Mult(u64),
    #[default]
    Square,
}

#[derive(Default, Debug)]
struct Monkey {
    items: VecDeque<u64>,
    op: Ops,
    test: u64,
    result: [usize; 2],
}

pub fn run() -> Result<()> {
    #[cfg(not(feature = "test_input"))]
    let file_contents = fs::read_to_string("day11.txt")?;

    #[cfg(feature = "test_input")]
    let file_contents = fs::read_to_string("tests/day11.txt")?;

    let mut monkeys: Vec<Monkey> = Vec::with_capacity(8);
    let mut lines = file_contents.lines();

    while let Some(line) = lines.next() {
        if line.is_empty() {
            continue;
        }
        if line.starts_with("Monkey") {
            let mut m = Monkey::default();
            for i in lines.next().unwrap().split_whitespace().skip(2) {
                if let Some(s) = i.strip_suffix(",") {
                    m.items.push_back(s.parse::<u64>().unwrap());
                } else {
                    m.items.push_back(i.parse::<u64>().unwrap());
                }
            }
            let mut ops = lines.next().unwrap().split_whitespace().skip(4);
            let op_symbol = ops.next().unwrap();
            let op_val = ops.next().unwrap();
            match op_symbol {
                "*" => {
                    m.op = if op_val == "old" {
                        Ops::Square
                    } else {
                        let val: u64 = op_val.parse().unwrap();
                        Ops::Mult(val)
                    }
                }
                "+" => {
                    let val: u64 = op_val.parse().unwrap();
                    m.op = Ops::Plus(val);
                }
                _ => return Err(anyhow!("invalid op code")),
            }
            let t = lines
                .next()
                .unwrap()
                .split_whitespace()
                .skip(3)
                .next()
                .unwrap()
                .parse::<u64>()
                .unwrap();
            m.test = t;
            let tmonkey: usize = lines
                .next()
                .unwrap()
                .split_whitespace()
                .skip(5)
                .next()
                .unwrap()
                .parse()
                .unwrap();
            let fmonkey: usize = lines
                .next()
                .unwrap()
                .split_whitespace()
                .skip(5)
                .next()
                .unwrap()
                .parse()
                .unwrap();
            m.result = [tmonkey, fmonkey];
            monkeys.push(m);
        }
    }

    part_one_items(&mut monkeys);
    part_two_items(&mut monkeys);

    Ok(())
}

fn match_op(op: &Ops, old: u64) -> u64 {
    match op {
        Ops::Plus(t) => old + t,
        Ops::Mult(t) => old * t,
        Ops::Square => old * old,
    }
}

fn part_one_items(monkeys: &mut Vec<Monkey>) {
    let mut inspections: Vec<usize> = Vec::with_capacity(8);
    let mut items: Vec<VecDeque<u64>> = monkeys.iter().map(|m| m.items.clone()).collect();

    for _ in 1..=20 {
        for m in 0..monkeys.len() {
            if let None = inspections.get(m) {
                inspections.push(0);
            }
            while let Some(item) = items[m].pop_front() {
                inspections[m] += 1;
                let score = (match_op(&monkeys[m].op, item) as f32 / 3.0).floor() as u64;
                if score % monkeys[m].test == 0 {
                    items[monkeys[m].result[0]].push_back(score);
                } else {
                    items[monkeys[m].result[1]].push_back(score);
                }
            }
        }
    }

    let mut bh: BinaryHeap<usize> = BinaryHeap::from(inspections);
    let one = bh.pop().unwrap();
    let two = bh.pop().unwrap();
    println!("part one {}*{}={}", one, two, one * two);
}

fn part_two_items(monkeys: &mut Vec<Monkey>) {
    let mut inspections: Vec<usize> = Vec::with_capacity(8);
    let mut items: Vec<VecDeque<u64>> = monkeys.iter().map(|m| m.items.clone()).collect();

    #[cfg(feature = "test_input")]
    let cd = 96577;

    // grep Test day11.txt | awk 'BEGIN { S=1} { S=S*$4} END {print S}'
    #[cfg(not(feature = "test_input"))]
    let cd = 9699690;

    for _ in 1..=10000 {
        for m in 0..monkeys.len() {
            if let None = inspections.get(m) {
                inspections.push(0);
            }
            while let Some(item) = items[m].pop_front() {
                inspections[m] += 1;
                let score = match_op(&monkeys[m].op, item);
                if score % monkeys[m].test == 0 {
                    items[monkeys[m].result[0]].push_back(score % cd);
                } else {
                    items[monkeys[m].result[1]].push_back(score % cd);
                }
            }
        }
    }

    let mut bh: BinaryHeap<usize> = BinaryHeap::from(inspections);
    let one = bh.pop().unwrap();
    let two = bh.pop().unwrap();
    println!("part two {}*{}={}", one, two, one * two);
}
