use anyhow::{anyhow, Result};
use std::fs;

#[derive(Clone, Copy)]
enum Tile {
    Air,
    Rock,
    Source,
    Sand,
}

type Cave = [[Tile; 800]; 200];

pub fn run() -> Result<()> {
    #[cfg(not(feature = "test_input"))]
    let file_contents = fs::read_to_string("day14.txt")?;

    #[cfg(feature = "test_input")]
    let file_contents = fs::read_to_string("tests/day14.txt")?;

    let mut cave: Cave = [[Tile::Air; 800]; 200];
    cave[0][500] = Tile::Source;
    let mut minx = 520;
    let mut maxx = 500;
    let mut maxy = 0;
    for line in file_contents.lines() {
        let mut lastx = 0;
        let mut lasty = 0;
        for p in line.split(" -> ").map(|e| {
            e.split(",")
                .map(|f| f.parse::<usize>().unwrap())
                .collect::<Vec<usize>>()
        }) {
            if minx > p[0] {
                minx = p[0];
            }
            if maxx < p[0] {
                maxx = p[0];
            }
            if maxy < p[1] {
                maxy = p[1];
            }

            if lastx == 0 && lasty == 0 {
                lastx = p[0];
                lasty = p[1];
                continue;
            }

            if lastx == p[0] {
                // vertical
                let r = if lasty < p[1] {
                    lasty..=p[1]
                } else {
                    p[1]..=lasty
                };
                for y in r {
                    cave[y][lastx] = Tile::Rock;
                }
            } else if lasty == p[1] {
                // horizontal
                let r = if lastx < p[0] {
                    lastx..=p[0]
                } else {
                    p[0]..=lastx
                };
                for x in r {
                    cave[lasty][x] = Tile::Rock;
                }
            } else {
                return Err(anyhow!("diagonal line"));
            }

            lastx = p[0];
            lasty = p[1];
        }
    }

    part1_sand_units(&cave, minx, maxx, maxy);
    part2_sand_units(&cave, maxy);

    Ok(())
}

fn print_cave(cave: &Cave, minx: usize, miny: usize, maxx: usize, maxy: usize) {
    println!("cave {},{} -> {},{}", minx, miny, maxx, maxy);
    for y in miny..=maxy {
        for x in minx..=maxx {
            match cave[y][x] {
                Tile::Air => print!("."),
                Tile::Rock => print!("#"),
                Tile::Source => print!("+"),
                Tile::Sand => print!("o"),
            }
        }
        println!();
    }
}

fn part1_sand_units(c: &Cave, minx: usize, maxx: usize, maxy: usize) {
    let mut cave = c.clone();
    let mut units = 0;

    loop {
        let mut x = 500;
        let mut placed = false;
        for i in 0..200 {
            if i > maxy || placed {
                break;
            }
            match cave[i + 1][x] {
                Tile::Air => {}
                Tile::Source => {}
                _ => {
                    if let Tile::Air = cave[i + 1][x - 1] {
                        x -= 1;
                    } else if let Tile::Air = cave[i + 1][x + 1] {
                        x += 1;
                    } else {
                        placed = true;
                        cave[i][x] = Tile::Sand;
                        break;
                    }
                }
            }
        }
        if placed {
            units += 1;
        } else {
            break;
        }
    }
    println!("part one units {}", units);
    print_cave(&cave, minx, 0, maxx, maxy);
}

fn part2_sand_units(c: &Cave, maxy: usize) {
    let mut cave = c.clone();
    let mut units = 0;
    let mut minx = 800;
    let mut maxx = 0;

    let floor = maxy + 2;
    for i in 0..800 {
        cave[floor][i] = Tile::Rock;
    }

    loop {
        if let Tile::Sand = cave[0][500] {
            break;
        }
        let mut x = 500;
        let mut placed = false;
        for i in 0..floor {
            if placed {
                if x < minx {
                    minx = x;
                }
                if x > maxx {
                    maxx = x;
                }
                break;
            }
            match cave[i + 1][x] {
                Tile::Air => {}
                Tile::Source => {}
                _ => {
                    if let Tile::Air = cave[i + 1][x - 1] {
                        x -= 1;
                    } else if let Tile::Air = cave[i + 1][x + 1] {
                        x += 1;
                    } else {
                        placed = true;
                        cave[i][x] = Tile::Sand;
                    }
                }
            }
        }
        if placed {
            units += 1;
        } else {
            break;
        }
    }
    println!("part two units {}", units);
    print_cave(&cave, minx, 0, maxx, floor);
}
